use std::io::BufRead;
use std::str::FromStr;

use quick_xml::events::BytesStart;
use quick_xml::Reader;

use crate::common::ShortString;
use crate::error::Error;
use crate::error::InvalidValue;
use crate::parser::utils::decode_attribute;
use crate::parser::utils::extract_attribute;
use crate::parser::utils::get_evidences;
use crate::parser::FromXml;

#[derive(Debug, Clone, Default)]
/// Describes the names for the protein and parts thereof.
pub struct Protein {
    pub name: Nomenclature,
    pub domains: Vec<Nomenclature>,
    pub components: Vec<Nomenclature>,
}

impl FromXml for Protein {
    fn from_xml<B: BufRead>(
        event: &BytesStart,
        reader: &mut Reader<B>,
        buffer: &mut Vec<u8>,
    ) -> Result<Self, Error> {
        let mut protein = Protein::default();
        parse_inner! {event, reader, buffer,
            e @ b"recommendedName" => {
                protein.name.recommended = FromXml::from_xml(&e, reader, buffer).map(Some)?;
            },
            e @ b"alternativeName" => {
                protein.name.alternative.push(FromXml::from_xml(&e, reader, buffer)?);
            },
            e @ b"submittedName" => {
                protein.name.submitted.push(FromXml::from_xml(&e, reader, buffer)?);
            },
            e @ b"component" => {
                // TODO: proper fix to avoid nested `component` in `component`
                protein.components.push(Self::from_xml(&e, reader, buffer)?.name);
            },
            e @ b"domain" => {
                // TODO: proper fix to avoid nested `domain` in `component`
                protein.domains.push(Self::from_xml(&e, reader, buffer)?.name);
            },
            e @ b"allergenName" => {
                let value = parse_text!(e, reader, buffer);
                if protein.name.allergen.replace(value).is_some() {
                    return Err(Error::DuplicateElement("allergen", "protein"));
                }
            },
            e @ b"biotechName" => {
                let value = parse_text!(e, reader, buffer);
                if protein.name.biotech.replace(value).is_some() {
                    return Err(Error::DuplicateElement("biotech", "protein"));
                }
            },
            e @ b"cdAntigenName" => {
                let value = parse_text!(e, reader, buffer);
                protein.name.cd_antigen.push(value);
            },
            e @ b"innName" => {
                let value = parse_text!(e, reader, buffer);
                protein.name.inn.push(value);

            }
        }

        Ok(protein)
    }
}

#[derive(Debug, Clone, Default)]
/// The different names that can be attached to a single protein.
pub struct Nomenclature {
    pub recommended: Option<Name>,
    pub alternative: Vec<Name>,
    pub submitted: Vec<Name>,
    pub allergen: Option<ShortString>, // FIXME: type should be EvidenceShortString?
    pub biotech: Option<ShortString>,
    pub cd_antigen: Vec<ShortString>,
    pub inn: Vec<ShortString>,
}

#[derive(Debug, Clone, Default)]
/// A single name in use for a protein.
pub struct Name {
    pub full: ShortString,
    pub short: Vec<ShortString>,
    pub ec_number: Vec<ShortString>,
}

impl FromXml for Name {
    fn from_xml<B: BufRead>(
        event: &BytesStart,
        reader: &mut Reader<B>,
        buffer: &mut Vec<u8>,
    ) -> Result<Self, Error> {
        let mut group = Self::default();

        parse_inner! {event, reader, buffer,
            e @ b"fullName" => {
                group.full = parse_text!(e, reader, buffer);
            },
            e @ b"shortName" => {
                group.short.push(parse_text!(e, reader, buffer));
            },
            e @ b"ecNumber" => {
                group.ec_number.push(parse_text!(e, reader, buffer));
            }
        };

        Ok(group)
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
/// The evidence supporting the existence of a protein.
pub enum ProteinExistence {
    ProteinLevelEvidence,
    TranscriptLevelEvidence,
    HomologyInferred,
    Predicted,
    Uncertain,
}

impl Default for ProteinExistence {
    fn default() -> Self {
        ProteinExistence::Uncertain
    }
}

impl FromStr for ProteinExistence {
    type Err = InvalidValue;
    fn from_str(s: &str) -> Result<Self, Self::Err> {
        use self::ProteinExistence::*;
        match s {
            "evidence at protein level" => Ok(ProteinLevelEvidence),
            "evidence at transcript level" => Ok(TranscriptLevelEvidence),
            "inferred from homology" => Ok(HomologyInferred),
            "predicted" => Ok(Predicted),
            "uncertain" => Ok(Uncertain),
            other => Err(InvalidValue::from(other)),
        }
    }
}

impl FromXml for ProteinExistence {
    fn from_xml<B: BufRead>(
        event: &BytesStart,
        reader: &mut Reader<B>,
        buffer: &mut Vec<u8>,
    ) -> Result<Self, Error> {
        debug_assert_eq!(event.local_name().as_ref(), b"proteinExistence");
        reader.read_to_end_into(event.name(), buffer)?;
        decode_attribute(event, reader, "type", "proteinExistence")
    }
}
